@@ -10,7 +10,9 @@ use {derive_deftly::Deftly, tor_rpcbase::templates::*};
 
 use crate::address::{IntoTorAddr, ResolveInstructions, StreamInstructions};
 
-use crate::config::{ClientAddrConfig, StreamTimeoutConfig, TorClientConfig};
+use crate::config::{
+    ClientAddrConfig, SoftwareStatusOverrideConfig, StreamTimeoutConfig, TorClientConfig,
+};
 use safelog::{sensitive, Sensitive};
 use tor_async_utils::{DropNotifyWatchSender, PostageWatchSenderExt};
 use tor_circmgr::isolation::{Isolation, StreamIsolation};
@@ -168,6 +170,8 @@ pub struct TorClient<R: Runtime> {
     addrcfg: Arc<MutCfg<ClientAddrConfig>>,
     /// Client DNS configuration
     timeoutcfg: Arc<MutCfg<StreamTimeoutConfig>>,
+    /// Software status configuration.
+    software_status_cfg: Arc<MutCfg<SoftwareStatusOverrideConfig>>,
     /// Mutex used to serialize concurrent attempts to reconfigure a TorClient.
     ///
     /// See [`TorClient::reconfigure`] for more information on its use.
@@ -471,6 +475,14 @@ impl InertTorClient {
             None => Ok(None),
         }
     }
+
+    /// Getter for keymgr.
+    #[cfg(feature = "onion-service-cli-extra")]
+    pub fn keymgr(&self) -> crate::Result<&KeyMgr> {
+        Ok(self.keymgr.as_ref().ok_or(ErrorDetail::KeystoreRequired {
+            action: "get key manager handle",
+        })?)
+    }
 }
 
 /// Preferences for whether a [`TorClient`] should bootstrap on its own or not.
@@ -782,6 +794,12 @@ impl TorClient<PreferredRuntime> {
     /// If using `async-std`, either take care to ensure Arti is not compiled with Tokio support,
     /// or manually create an `async-std` runtime using [`tor_rtcompat`] and use it with
     /// [`TorClient::with_runtime`].
+    ///
+    /// # Do not fork
+    ///
+    /// The process [**may not fork**](tor_rtcompat#do-not-fork)
+    /// (except, very carefully, before exec)
+    /// after calling this function, because it creates a [`PreferredRuntime`].
     pub async fn create_bootstrapped(config: TorClientConfig) -> crate::Result<Self> {
         let runtime = PreferredRuntime::current()
             .expect("TorClient could not get an asynchronous runtime; are you running in the right context?");
@@ -806,6 +824,12 @@ impl TorClient<PreferredRuntime> {
     /// If using `async-std`, either take care to ensure Arti is not compiled with Tokio support,
     /// or manually create an `async-std` runtime using [`tor_rtcompat`] and use it with
     /// [`TorClient::with_runtime`].
+    ///
+    /// # Do not fork
+    ///
+    /// The process [**may not fork**](tor_rtcompat#do-not-fork)
+    /// (except, very carefully, before exec)
+    /// after calling this function, because it creates a [`PreferredRuntime`].
     pub fn builder() -> TorClientBuilder<PreferredRuntime> {
         let runtime = PreferredRuntime::current()
             .expect("TorClient could not get an asynchronous runtime; are you running in the right context?");
@@ -916,6 +940,34 @@ impl<R: Runtime> TorClient<R> {
             )
             .map_err(crate::Error::into_detail)?;
 
+        let software_status_cfg = Arc::new(MutCfg::new(config.use_obsolete_software.clone()));
+        let rtclone = runtime.clone();
+        #[allow(clippy::print_stderr)]
+        crate::protostatus::enforce_protocol_recommendations(
+            &runtime,
+            Arc::clone(&dirmgr),
+            crate::software_release_date(),
+            crate::supported_protocols(),
+            Arc::clone(&software_status_cfg),
+            // TODO #1932: It would be nice to have a cleaner shutdown mechanism here,
+            // but that will take some work.
+            |fatal| async move {
+                use tor_error::ErrorReport as _;
+                // We already logged this error, but let's tell stderr too.
+                eprintln!(
+                    "Shutting down because of unsupported software version.\nError was:\n{}",
+                    fatal.report(),
+                );
+                if let Some(hint) = crate::err::Error::from(fatal).hint() {
+                    eprintln!("{}", hint);
+                }
+                // Give the tracing module a while to flush everything, since it has no built-in
+                // flush function.
+                rtclone.sleep(std::time::Duration::new(5, 0)).await;
+                std::process::exit(1);
+            },
+        )?;
+
         let mut periodic_task_handles = circmgr
             .launch_background_tasks(&runtime, &dirmgr, statemgr.clone())
             .map_err(ErrorDetail::CircMgrSetup)?;
@@ -1012,6 +1064,7 @@ impl<R: Runtime> TorClient<R> {
             #[cfg(feature = "onion-service-service")]
             state_directory,
             path_resolver,
+            software_status_cfg,
         })
     }
 
@@ -1245,6 +1298,8 @@ impl<R: Runtime> TorClient<R> {
 
         self.addrcfg.replace(addr_cfg.clone());
         self.timeoutcfg.replace(timeout_cfg.clone());
+        self.software_status_cfg
+            .replace(new_config.use_obsolete_software.clone());
 
         Ok(())
     }
@@ -1380,6 +1435,8 @@ impl<R: Runtime> TorClient<R> {
                 hostname,
                 port,
             } => {
+                use safelog::DisplayRedacted as _;
+
                 self.wait_for_bootstrap().await?;
                 let netdir = self.netdir(Timeliness::Timely, "connect to a hidden service")?;
 
@@ -1388,15 +1445,14 @@ impl<R: Runtime> TorClient<R> {
                 if let Some(keymgr) = &self.inert_client.keymgr {
                     let desc_enc_key_spec = HsClientDescEncKeypairSpecifier::new(hsid);
 
-                    // TODO hs: refactor to reduce code duplication.
-                    //
-                    // The code that reads ks_hsc_desc_enc and ks_hsc_intro_auth and builds the
-                    // HsClientSecretKeys is very repetitive and should be refactored.
                     let ks_hsc_desc_enc =
                         keymgr.get::<HsClientDescEncKeypair>(&desc_enc_key_spec)?;
 
                     if let Some(ks_hsc_desc_enc) = ks_hsc_desc_enc {
-                        debug!("Found descriptor decryption key for {hsid}");
+                        debug!(
+                            "Found descriptor decryption key for {}",
+                            hsid.display_redacted()
+                        );
                         hs_client_secret_keys_builder.ks_hsc_desc_enc(ks_hsc_desc_enc);
                     }
                 };
@@ -1414,10 +1470,7 @@ impl<R: Runtime> TorClient<R> {
                         self.isolation(prefs),
                     )
                     .await
-                    .map_err(|cause| ErrorDetail::ObtainHsCircuit {
-                        cause,
-                        hsid: hsid.into(),
-                    })?;
+                    .map_err(|cause| ErrorDetail::ObtainHsCircuit { cause, hsid })?;
                 // On connections to onion services, we have to suppress
                 // everything except the port from the BEGIN message.  We also
                 // disable optimistic data.
@@ -1689,7 +1742,6 @@ impl<R: Runtime> TorClient<R> {
                 action: "launch onion service",
             })?
             .clone();
-        // NOTE: Intended usage is via StateDirectory::acquire_instance(), but OnionServiceBuilder requires ownership.
         let state_dir = self.state_directory.clone();
 
         let service = tor_hsservice::OnionService::builder()
@@ -1965,8 +2017,7 @@ impl<R: Runtime> TorClient<R> {
             .keymgr(keymgr)
             .state_dir(state_dir)
             .build()
-            // TODO: do we need an ErrorDetail::CreateOnionService?
-            .map_err(ErrorDetail::LaunchOnionService)?)
+            .map_err(ErrorDetail::OnionServiceSetup)?)
     }
 
     /// Return a current [`status::BootstrapStatus`] describing how close this client
