@@ -41,6 +41,7 @@
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
 #![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 use std::collections::BinaryHeap;
@@ -594,7 +595,7 @@ macro_rules! derive_serde_raw { {
 /// Flatten a `Result<Result<T, E>, E>` into a `Result<T, E>`.
 ///
 /// See the nightly [`Result::flatten`].
-// TODO: When `Result::flatten` is stable and our MSRV allows,
+// TODO MSRV TBD: When `Result::flatten` is stable and our MSRV allows,
 // remove this function and replace uses with `Result::flatten`.
 pub fn flatten<T, E>(x: Result<Result<T, E>, E>) -> Result<T, E> {
     match x {
@@ -605,6 +606,25 @@ pub fn flatten<T, E>(x: Result<Result<T, E>, E>) -> Result<T, E> {
 
 // ----------------------------------------------------------------------
 
+/// Asserts that the type of the expression implements the given trait.
+///
+/// Example:
+///
+/// ```
+/// # use tor_basic_utils::assert_val_impl_trait;
+/// let x: u32 = 0;
+/// assert_val_impl_trait!(x, Clone);
+/// ```
+#[macro_export]
+macro_rules! assert_val_impl_trait {
+    ($check:expr, $trait:path $(,)?) => {{
+        fn ensure_trait<T: $trait>(_s: &T) {}
+        ensure_trait(&$check);
+    }};
+}
+
+// ----------------------------------------------------------------------
+
 #[cfg(test)]
 mod test {
     // @@ begin test lint list maintained by maint/add_warning @@
