@@ -29,7 +29,7 @@ use tracing::{debug, info, trace, warn};
 
 use crate::storage::Store;
 #[cfg(test)]
-use once_cell::sync::Lazy;
+use std::sync::LazyLock;
 #[cfg(test)]
 use std::sync::Mutex;
 use tor_circmgr::{CircMgr, DirInfo};
@@ -297,12 +297,13 @@ async fn fetch_single<R: Runtime>(
 /// Note that only one test uses this: otherwise there would be a race
 /// condition. :p
 #[cfg(test)]
-static CANNED_RESPONSE: Lazy<Mutex<Vec<String>>> = Lazy::new(|| Mutex::new(vec![]));
+static CANNED_RESPONSE: LazyLock<Mutex<Vec<String>>> = LazyLock::new(|| Mutex::new(vec![]));
 
 /// Launch a set of download requests for a set of missing objects in
 /// `missing`, and return each request along with the response it received.
 ///
 /// Don't launch more than `parallelism` requests at once.
+#[allow(clippy::cognitive_complexity)] // TODO: maybe refactor?
 async fn fetch_multiple<R: Runtime>(
     dirmgr: Arc<DirMgr<R>>,
     attempt_id: AttemptId,
@@ -399,6 +400,7 @@ async fn load_once<R: Runtime>(
 /// cache in `dirmgr`, advancing the state to the extent possible.
 ///
 /// No downloads are performed; the provided state will not be reset.
+#[allow(clippy::cognitive_complexity)] // TODO: Refactor? Somewhat due to tracing.
 pub(crate) async fn load<R: Runtime>(
     dirmgr: Arc<DirMgr<R>>,
     mut state: Box<dyn DirState>,
@@ -454,6 +456,7 @@ pub(crate) async fn load<R: Runtime>(
 ///
 /// This can launch one or more download requests, but will not launch more
 /// than `parallelism` requests at a time.
+#[allow(clippy::cognitive_complexity)] // TODO: Refactor?
 async fn download_attempt<R: Runtime>(
     dirmgr: &Arc<DirMgr<R>>,
     state: &mut Box<dyn DirState>,
@@ -537,6 +540,7 @@ async fn download_attempt<R: Runtime>(
 ///
 /// The first time that the state becomes ["usable"](Readiness::Usable), notify
 /// the sender in `on_usable`.
+#[allow(clippy::cognitive_complexity)] // TODO: Refactor!
 pub(crate) async fn download<R: Runtime>(
     dirmgr: Weak<DirMgr<R>>,
     state: &mut Box<dyn DirState>,
@@ -573,12 +577,6 @@ pub(crate) async fn download<R: Runtime>(
             dirmgr.runtime.wallclock()
         };
 
-        // Skip the downloads if we can...
-        if state.can_advance() {
-            advance(state);
-            trace!(attempt=%attempt_id, state=%state.describe(), "State has advanced.");
-            continue 'next_state;
-        }
         // Apply any netdir changes that the state gives us.
         // TODO(eta): Consider deprecating state.is_ready().
         {
@@ -587,6 +585,12 @@ pub(crate) async fn download<R: Runtime>(
             dirmgr.apply_netdir_changes(state, &mut **store)?;
             dirmgr.update_progress(attempt_id, state.bootstrap_progress());
         }
+        // Skip the downloads if we can...
+        if state.can_advance() {
+            advance(state);
+            trace!(attempt=%attempt_id, state=%state.describe(), "State has advanced.");
+            continue 'next_state;
+        }
         if state.is_ready(Readiness::Complete) {
             trace!(attempt=%attempt_id, state=%state.describe(), "Directory is now Complete.");
             return Ok(());
