@@ -434,7 +434,7 @@ impl Ipt {
                     return Err(FatalError::IptKeysFoundUnexpectedly(arti_path()?).into())
                 },
                 (Some(_), None) => {
-                    error!("HS service {} missing previous key {:?}, regenerating",
+                    error!("bug: HS service {} missing previous key {:?}. Regenerating.",
                            &imm.nick, arti_path()?);
                 }
              }
@@ -1440,6 +1440,7 @@ impl<R: Runtime, M: Mockable<R>> IptManager<R, M> {
     /// (Note that the number of IPTs can be significantly larger than
     /// the maximum target of 20, if the service is very busy so the intro points
     /// are cycling rapidly due to the need to replace the replay database.)
+    #[allow(clippy::cognitive_complexity)] // TODO: Refactor?
     async fn run_once(
         &mut self,
         // This is a separate argument for borrowck reasons
@@ -1576,7 +1577,7 @@ impl<R: Runtime, M: Mockable<R>> IptManager<R, M> {
             .await
             {
                 Err(crash) => {
-                    error!("HS service {} crashed! {}", &self.imm.nick, crash);
+                    error!("bug: HS service {} crashed! {}", &self.imm.nick, crash);
 
                     self.imm.status_tx.send_broken(crash);
                     break;
