@@ -41,6 +41,7 @@
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
 #![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 // TODO #1645 (either remove this, or decide to have it everywhere)
@@ -63,6 +64,7 @@ mod ipt_lid;
 mod ipt_mgr;
 mod ipt_set;
 mod keys;
+mod pow;
 mod publish;
 mod rend_handshake;
 mod replay;
@@ -88,6 +90,8 @@ pub mod time_store_for_doctests_unstable_no_semver_guarantees {
     pub use crate::time_store::*;
 }
 
+use std::pin::Pin;
+
 use internal_prelude::*;
 
 // ---------- public exports ----------
@@ -100,6 +104,7 @@ pub use keys::{
     BlindIdKeypairSpecifier, BlindIdPublicKeySpecifier, DescSigningKeypairSpecifier,
     HsIdKeypairSpecifier, HsIdPublicKeySpecifier,
 };
+use pow::{NewPowManager, PowManager};
 pub use publish::UploadError as DescUploadError;
 pub use req::{RendRequest, StreamRequest};
 pub use tor_hscrypto::pk::HsId;
@@ -147,8 +152,9 @@ struct SvcInner {
     status_tx: StatusSender,
 
     /// Handles that we'll take ownership of when launching the service.
+    #[allow(clippy::type_complexity)]
     unlaunched: Option<(
-        mpsc::Receiver<RendRequest>,
+        Pin<Box<dyn Stream<Item = RendRequest> + Send + Sync>>,
         Box<dyn Launchable + Send + Sync>,
     )>,
 }
@@ -173,6 +179,9 @@ struct ForLaunch<R: Runtime> {
     ///
     ///
     ipt_mgr_view: IptsManagerView,
+
+    /// Proof-of-work manager.
+    pow_manager: Arc<PowManager<R>>,
 }
 
 /// Private trait used to type-erase `ForLaunch<R>`, so that we don't need to
@@ -186,6 +195,7 @@ impl<R: Runtime> Launchable for ForLaunch<R> {
     fn launch(self: Box<Self>) -> Result<(), StartupError> {
         self.ipt_mgr.launch_background_tasks(self.ipt_mgr_view)?;
         self.publisher.launch()?;
+        self.pow_manager.launch()?;
 
         Ok(())
     }
@@ -292,9 +302,24 @@ impl OnionService {
             .storage_handle("iptpub")
             .map_err(StartupError::StateDirectoryInaccessible)?;
 
-        // If the HS implementation is stalled somehow, this is a local problem.
-        // We shouldn't kill the HS even if this is the oldest data in the system.
-        let (rend_req_tx, rend_req_rx) = mpsc_channel_no_memquota(32);
+        let pow_manager_storage_handle = state_handle
+            .storage_handle("pow_manager")
+            .map_err(StartupError::StateDirectoryInaccessible)?;
+        let pow_nonce_dir = state_handle
+            .raw_subdir("pow_nonces")
+            .map_err(StartupError::StateDirectoryInaccessible)?;
+        let NewPowManager {
+            pow_manager,
+            rend_req_tx,
+            rend_req_rx,
+            publisher_update_rx,
+        } = PowManager::new(
+            runtime.clone(),
+            nickname.clone(),
+            pow_nonce_dir,
+            keymgr.clone(),
+            pow_manager_storage_handle,
+        )?;
 
         let (shutdown_tx, shutdown_rx) = broadcast::channel(0);
         let (config_tx, config_rx) = postage::watch::channel_with(Arc::new(config));
@@ -329,6 +354,8 @@ impl OnionService {
             status_tx.clone().into(),
             Arc::clone(&keymgr),
             path_resolver,
+            pow_manager.clone(),
+            publisher_update_rx,
         );
 
         let svc = Arc::new(RunningOnionService {
@@ -344,6 +371,7 @@ impl OnionService {
                         publisher,
                         ipt_mgr,
                         ipt_mgr_view,
+                        pow_manager,
                     }),
                 )),
             }),
@@ -562,14 +590,14 @@ fn maybe_generate_hsid(
     if generated {
         info!(
             "Generated a new identity for service {nickname}: {}",
-            sensitive(hsid)
+            hsid.display_redacted()
         );
     } else {
         // TODO: We may want to downgrade this to trace once we have a CLI
         // for extracting it.
         info!(
             "Using existing identity for service {nickname}: {}",
-            sensitive(hsid)
+            hsid.display_redacted()
         );
     }
 
@@ -596,6 +624,23 @@ fn onion_address(keymgr: &KeyMgr, nickname: &HsNickname) -> Option<HsId> {
         .map(|hsid| hsid.id())
 }
 
+/// Return a list of the protocols[supported](tor_protover::doc_supported)
+/// by this crate, running as a hidden service.
+pub fn supported_hsservice_protocols() -> tor_protover::Protocols {
+    use tor_protover::named::*;
+    // WARNING: REMOVING ELEMENTS FROM THIS LIST CAN BE DANGEROUS!
+    // SEE [`tor_protover::doc_changing`]
+    [
+        //
+        HSINTRO_V3,
+        HSINTRO_RATELIM,
+        HSREND_V3,
+        HSDIR_V3,
+    ]
+    .into_iter()
+    .collect()
+}
+
 #[cfg(test)]
 pub(crate) mod test {
     // @@ begin test lint list maintained by maint/add_warning @@
@@ -631,6 +676,13 @@ pub(crate) mod test {
     /// The nickname of the test service.
     const TEST_SVC_NICKNAME: &str = "test-svc";
 
+    #[test]
+    fn protocols() {
+        let pr = supported_hsservice_protocols();
+        let expected = "HSIntro=4-5 HSRend=2 HSDir=2".parse().unwrap();
+        assert_eq!(pr, expected);
+    }
+
     /// Make a fresh `KeyMgr` (containing no keys) using files in `temp_dir`
     pub(crate) fn create_keymgr(temp_dir: &TestTempDir) -> TestTempDirGuard<Arc<KeyMgr>> {
         temp_dir.subdir_used_by("keystore", |keystore_dir| {
