@@ -13,6 +13,8 @@
 //! a file on disk.
 
 mod ipt;
+#[cfg(feature = "hs-pow-full")]
+mod pow;
 
 use crate::internal_prelude::*;
 
@@ -44,6 +46,10 @@ pub(crate) struct ReplayLog<T> {
 /// A [`ReplayLog`] for [`Introduce2`](tor_cell::relaycell::msg::Introduce2) messages.
 pub(crate) type IptReplayLog = ReplayLog<ipt::IptReplayLogType>;
 
+/// A [`ReplayLog`] for Proof-of-Work [`Nonce`](tor_hscrypto::pow::v1::Nonce)s.
+#[cfg(feature = "hs-pow-full")]
+pub(crate) type PowNonceReplayLog = ReplayLog<pow::PowNonceReplayLogType>;
+
 /// The length of the [`ReplayLogType::MAGIC`] constant.
 ///
 // TODO: If Rust's constant expressions supported generics we wouldn't need this at all.
@@ -649,7 +655,7 @@ mod test {
             }
 
             let demand_efbig = |e| match e {
-                // MSRV:: io::ErrorKind::FileTooLarge is still unstable
+                // TODO MSRV 1.83: replace with io::ErrorKind::FileTooLarge?
                 ReplayError::Log(e) if e.raw_os_error() == Some(libc::EFBIG) => {}
                 other => panic!("expected EFBUG, got {other:?}"),
             };
