@@ -100,11 +100,8 @@ pub enum WeightRole {
     Unweighted,
     /// Selecting a relay for use as a hidden service introduction point
     HsIntro,
-    // Note: There is no `HsRend` role, since in practice when we want to pick a
-    // rendezvous point we use a pre-built circuit from our circuit-pool, the
-    // last hop of which was selected with the `Middle` weight.  Fortunately,
-    // the weighting rules for picking rendezvous points are the same as for
-    // picking middle relays.
+    /// Selecting a relay for use as a hidden service rendezvous point
+    HsRend,
 }
 
 /// Description for how to weight a single kind of relay for each WeightRole.
@@ -163,6 +160,7 @@ impl RelayWeight {
             WeightRole::Exit => self.as_exit,
             WeightRole::BeginDir => self.as_dir,
             WeightRole::HsIntro => self.as_middle, // TODO SPEC is this right?
+            WeightRole::HsRend => self.as_middle,  // TODO SPEC is this right?
             WeightRole::Unweighted => 1,
         }
     }
@@ -400,11 +398,7 @@ fn clamp_to_pos(inp: i32) -> u32 {
 /// inside 64 bits.
 fn calculate_shift(a: u64, b: u64) -> u32 {
     let bits_for_product = log2_upper(a) + log2_upper(b);
-    if bits_for_product < 64 {
-        0
-    } else {
-        bits_for_product - 64
-    }
+    bits_for_product.saturating_sub(64)
 }
 
 /// Return an upper bound for the log2 of n.
