@@ -11,8 +11,8 @@ use crate::types::misc::{UnvalidatedEdCert, B64};
 use crate::{NetdocErrorKind as EK, Result};
 
 use itertools::Itertools as _;
-use once_cell::sync::Lazy;
 use smallvec::SmallVec;
+use std::sync::LazyLock;
 use tor_checkable::signed::SignatureGated;
 use tor_checkable::timed::TimerangeBound;
 use tor_checkable::Timebound;
@@ -63,7 +63,7 @@ decl_keyword! {
 
 /// Rules about how keywords appear in the header part of an onion service
 /// descriptor.
-static HS_INNER_HEADER_RULES: Lazy<SectionRules<HsInnerKwd>> = Lazy::new(|| {
+static HS_INNER_HEADER_RULES: LazyLock<SectionRules<HsInnerKwd>> = LazyLock::new(|| {
     use HsInnerKwd::*;
 
     let mut rules = SectionRules::builder();
@@ -78,7 +78,7 @@ static HS_INNER_HEADER_RULES: Lazy<SectionRules<HsInnerKwd>> = Lazy::new(|| {
 
 /// Rules about how keywords appear in each introduction-point section of an
 /// onion service descriptor.
-static HS_INNER_INTRO_RULES: Lazy<SectionRules<HsInnerKwd>> = Lazy::new(|| {
+static HS_INNER_INTRO_RULES: LazyLock<SectionRules<HsInnerKwd>> = LazyLock::new(|| {
     use HsInnerKwd::*;
 
     let mut rules = SectionRules::builder();
@@ -496,7 +496,8 @@ mod test {
                 .1,
         );
         for n in NUM_INTRO_POINT_MAX..NUM_INTRO_POINT_MAX + 2 {
-            let many = chain!(iter::once(&*none), iter::repeat(&*ipt).take(n),).collect::<String>();
+            let many =
+                chain!(iter::once(&*none), std::iter::repeat_n(&*ipt, n),).collect::<String>();
             let desc = HsDescInner::parse(&many).unwrap();
             let desc = desc
                 .1
