@@ -535,7 +535,6 @@ mod test {
     #![allow(clippy::needless_pass_by_value)]
     //! <!-- @@ end test lint list maintained by maint/add_warning @@ -->
     use itertools::Itertools;
-    use std::iter;
 
     use base64ct::Encoding;
 
@@ -617,10 +616,7 @@ mod test {
         // Check that strings that we accept are precisely ones which
         // can be generated by either Base64 or Base64Unpadded
         for n in 0..=5 {
-            for c_vec in iter::repeat("ACEQg/=".chars())
-                .take(n)
-                .multi_cartesian_product()
-            {
+            for c_vec in std::iter::repeat_n("ACEQg/=".chars(), n).multi_cartesian_product() {
                 let s: String = c_vec.into_iter().collect();
                 #[allow(clippy::print_stderr)]
                 let b = match s.parse::<B64>() {
