@@ -355,8 +355,8 @@ impl Sink<AnyChanCell> for ChannelSender {
             match cell.msg() {
                 Relay(_) | Padding(_) | Vpadding(_) => {} // too frequent to log.
                 _ => trace!(
-                    "{}: Sending {} for {}",
-                    this.unique_id,
+                    channel_id = %this.unique_id,
+                    "Sending {} for {}",
                     cell.msg().cmd(),
                     CircId::get_or_zero(cell.circid())
                 ),
@@ -684,7 +684,8 @@ impl Channel {
         let memquota = CircuitAccount::new(&self.details.memquota)?;
 
         // TODO: blocking is risky, but so is unbounded.
-        let (sender, receiver) = MpscSpec::new(128).new_mq(time_prov, memquota.as_raw_account())?;
+        let (sender, receiver) =
+            MpscSpec::new(128).new_mq(time_prov.clone(), memquota.as_raw_account())?;
         let (createdsender, createdreceiver) = oneshot::channel::<CreateResponse>();
 
         let (tx, rx) = oneshot::channel();
@@ -703,6 +704,7 @@ impl Channel {
             createdreceiver,
             receiver,
             circ_unique_id,
+            time_prov,
             memquota,
         ))
     }
