@@ -17,8 +17,10 @@ mod flow_control;
 #[cfg(feature = "hs-service")]
 mod incoming;
 mod params;
+pub(crate) mod queue;
 mod raw;
 mod resolve;
+pub(crate) mod xon_xoff;
 
 pub(crate) use cmdcheck::{AnyCmdChecker, CmdChecker, StreamStatus};
 pub use data::{DataReader, DataStream, DataWriter};
@@ -32,7 +34,7 @@ pub use incoming::{
     IncomingStreamRequestDisposition, IncomingStreamRequestFilter,
 };
 pub use params::StreamParameters;
-pub use raw::StreamReader;
+pub use raw::StreamReceiver;
 pub use resolve::ResolveStream;
 pub(crate) use {data::DataCmdChecker, resolve::ResolveCmdChecker};
 
@@ -42,4 +44,4 @@ pub use tor_cell::relaycell::msg::IpVersionPreference;
 #[cfg_attr(docsrs, doc(cfg(feature = "stream-ctrl")))]
 pub use {ctrl::ClientStreamCtrl, data::ClientDataStreamCtrl};
 
-pub(crate) use flow_control::StreamSendFlowControl;
+pub(crate) use flow_control::{DrainRateRequest, StreamFlowControl, StreamRateLimit};
