@@ -1,8 +1,9 @@
 //! Types and code for mapping StreamIDs to streams on a circuit.
 
 use crate::congestion::sendme;
-use crate::stream::{AnyCmdChecker, StreamSendFlowControl};
-use crate::tunnel::circuit::{StreamMpscReceiver, StreamMpscSender};
+use crate::stream::queue::StreamQueueSender;
+use crate::stream::{AnyCmdChecker, StreamFlowControl};
+use crate::tunnel::circuit::StreamMpscReceiver;
 use crate::tunnel::halfstream::HalfStream;
 use crate::tunnel::reactor::circuit::RECV_WINDOW_INIT;
 use crate::util::stream_poll_set::{KeyAlreadyInsertedError, StreamPollSet};
@@ -10,6 +11,7 @@ use crate::{Error, Result};
 use pin_project::pin_project;
 use tor_async_utils::peekable_stream::{PeekableStream, UnobtrusivePeekableStream};
 use tor_async_utils::stream_peek::StreamUnobtrusivePeeker;
+use tor_cell::relaycell::flow_ctrl::{Xoff, Xon, XonKbpsEwma};
 use tor_cell::relaycell::{msg::AnyRelayMsg, StreamId};
 use tor_cell::relaycell::{RelayMsg, UnparsedRelayMsg};
 
@@ -32,7 +34,7 @@ use tracing::debug;
 #[pin_project]
 pub(super) struct OpenStreamEnt {
     /// Sink to send relay cells tagged for this stream into.
-    pub(super) sink: StreamMpscSender<UnparsedRelayMsg>,
+    pub(super) sink: StreamQueueSender,
     /// Number of cells dropped due to the stream disappearing before we can
     /// transform this into an `EndSent`.
     pub(super) dropped: u16,
@@ -41,7 +43,7 @@ pub(super) struct OpenStreamEnt {
     /// Flow control for this stream.
     // Non-pub because we need to proxy `put_for_incoming_sendme` to ensure
     // `flow_ctrl_waker` is woken.
-    flow_ctrl: StreamSendFlowControl,
+    flow_ctrl: StreamFlowControl,
     /// Stream for cells that should be sent down this stream.
     // Not directly exposed. This should only be polled via
     // `OpenStreamEntStream`s implementation of `Stream`, which in turn should
@@ -61,12 +63,10 @@ impl OpenStreamEnt {
 
     /// Handle an incoming sendme.
     ///
-    /// On success, return the number of cells left in the window.
-    ///
     /// On failure, return an error: the caller should close the stream or
     /// circuit with a protocol error.
-    pub(crate) fn put_for_incoming_sendme(&mut self) -> Result<()> {
-        self.flow_ctrl.put_for_incoming_sendme()?;
+    pub(crate) fn put_for_incoming_sendme(&mut self, msg: UnparsedRelayMsg) -> Result<()> {
+        self.flow_ctrl.put_for_incoming_sendme(msg)?;
         // Wake the stream if it was blocked on flow control.
         if let Some(waker) = self.flow_ctrl_waker.take() {
             waker.wake();
@@ -74,6 +74,56 @@ impl OpenStreamEnt {
         Ok(())
     }
 
+    /// The approximate number of stream inbound data bytes buffered.
+    fn approx_stream_bytes_buffered(&self) -> usize {
+        // NOTE: Here we want to know the total number of buffered incoming stream data bytes. We
+        // have access to the `StreamQueueSender` and can get how many bytes are buffered in that
+        // queue.
+        // But this isn't always the total number of buffered bytes since some bytes might be
+        // buffered outside of this queue.
+        // For example `DataReaderImpl` stores some stream bytes in its `pending` buffer, and we
+        // have no way to access that from here in the reactor. So it's impossible to know the total
+        // number of incoming stream data bytes that are buffered.
+        //
+        // This isn't really an issue in practice since *most* of the bytes will be queued in the
+        // `StreamQueueSender`, the XOFF threshold is very large, and we don't need to be exact.
+        self.sink.approx_stream_bytes()
+    }
+
+    /// Check if we should send an XON message.
+    ///
+    /// If we should, then returns the XON message that should be sent.
+    /// Returns an error if XON/XOFF messages aren't supported for this type of flow control.
+    pub(crate) fn maybe_send_xon(&mut self, rate: XonKbpsEwma) -> Result<Option<Xon>> {
+        self.flow_ctrl
+            .maybe_send_xon(rate, self.approx_stream_bytes_buffered())
+    }
+
+    /// Check if we should send an XOFF message.
+    ///
+    /// If we should, then returns the XOFF message that should be sent.
+    /// Returns an error if XON/XOFF messages aren't supported for this type of flow control.
+    pub(super) fn maybe_send_xoff(&mut self) -> Result<Option<Xoff>> {
+        self.flow_ctrl
+            .maybe_send_xoff(self.approx_stream_bytes_buffered())
+    }
+
+    /// Handle an incoming XON message.
+    ///
+    /// On failure, return an error: the caller should close the stream or
+    /// circuit with a protocol error.
+    pub(crate) fn handle_incoming_xon(&mut self, msg: UnparsedRelayMsg) -> Result<()> {
+        self.flow_ctrl.handle_incoming_xon(msg)
+    }
+
+    /// Handle an incoming XOFF message.
+    ///
+    /// On failure, return an error: the caller should close the stream or
+    /// circuit with a protocol error.
+    pub(crate) fn handle_incoming_xoff(&mut self, msg: UnparsedRelayMsg) -> Result<()> {
+        self.flow_ctrl.handle_incoming_xoff(msg)
+    }
+
     /// Take capacity to send `msg`. If there's insufficient capacity, returns
     /// an error. Should be called at the point we've fully committed to
     /// sending the message.
@@ -269,15 +319,15 @@ impl StreamMap {
     /// Add an entry to this map; return the newly allocated StreamId.
     pub(super) fn add_ent(
         &mut self,
-        sink: StreamMpscSender<UnparsedRelayMsg>,
+        sink: StreamQueueSender,
         rx: StreamMpscReceiver<AnyRelayMsg>,
-        send_window: sendme::StreamSendWindow,
+        flow_ctrl: StreamFlowControl,
         cmd_checker: AnyCmdChecker,
     ) -> Result<StreamId> {
         let mut stream_ent = OpenStreamEntStream {
             inner: OpenStreamEnt {
                 sink,
-                flow_ctrl: StreamSendFlowControl::new_window_based(send_window),
+                flow_ctrl,
                 dropped: 0,
                 cmd_checker,
                 rx: StreamUnobtrusivePeeker::new(rx),
@@ -309,16 +359,16 @@ impl StreamMap {
     #[cfg(feature = "hs-service")]
     pub(super) fn add_ent_with_id(
         &mut self,
-        sink: StreamMpscSender<UnparsedRelayMsg>,
+        sink: StreamQueueSender,
         rx: StreamMpscReceiver<AnyRelayMsg>,
-        send_window: sendme::StreamSendWindow,
+        flow_ctrl: StreamFlowControl,
         id: StreamId,
         cmd_checker: AnyCmdChecker,
     ) -> Result<()> {
         let stream_ent = OpenStreamEntStream {
             inner: OpenStreamEnt {
                 sink,
-                flow_ctrl: StreamSendFlowControl::new_window_based(send_window),
+                flow_ctrl,
                 dropped: 0,
                 cmd_checker,
                 rx: StreamUnobtrusivePeeker::new(rx),
@@ -514,6 +564,7 @@ mod test {
     #![allow(clippy::needless_pass_by_value)]
     //! <!-- @@ end test lint list maintained by maint/add_warning @@ -->
     use super::*;
+    use crate::stream::queue::fake_stream_queue;
     use crate::tunnel::circuit::test::fake_mpsc;
     use crate::{congestion::sendme::StreamSendWindow, stream::DataCmdChecker};
 
@@ -537,12 +588,12 @@ mod test {
 
         // Try add_ent
         for n in 1..=128 {
-            let (sink, _) = fake_mpsc(128);
+            let (sink, _) = fake_stream_queue(128);
             let (_, rx) = fake_mpsc(2);
             let id = map.add_ent(
                 sink,
                 rx,
-                StreamSendWindow::new(500),
+                StreamFlowControl::new_window_based(StreamSendWindow::new(500)),
                 DataCmdChecker::new_any(),
             )?;
             let expect_id: StreamId = next_id;
