@@ -41,6 +41,7 @@
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
 #![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 // TODO #1645 (either remove this, or decide to have it everywhere)
@@ -111,6 +112,10 @@ use tokio as preferred_backend_mod;
 /// performance.
 /// If `native_tls` and `rustls` are both available, we prefer `native_tls` since
 /// it has been used in Arti for longer.
+///
+/// The process [**may not fork**](crate#do-not-fork)
+/// (except, very carefully, before exec)
+/// after creating this or any other `Runtime`.
 #[cfg(all(
     any(feature = "native-tls", feature = "rustls"),
     any(feature = "async-std", feature = "tokio")
@@ -158,6 +163,9 @@ impl PreferredRuntime {
     /// If the `tor-rtcompat` crate was compiled with `tokio` support,
     /// this function will never return a runtime based on `async_std`.
     ///
+    /// The process [**may not fork**](crate#do-not-fork)
+    /// (except, very carefully, before exec)
+    /// after creating this or any other `Runtime`.
     //
     // ## Note to Arti developers
     //
@@ -183,6 +191,10 @@ impl PreferredRuntime {
     ///
     /// If you need more fine-grained control over a runtime, you can create it
     /// using an appropriate builder type or function.
+    ///
+    /// The process [**may not fork**](crate#do-not-fork)
+    /// (except, very carefully, before exec)
+    /// after creating this or any other `Runtime`.
     //
     // ## Note to Arti developers
     //
