@@ -9,20 +9,18 @@ use std::fmt::{self, Debug, Display};
 use std::future::Future;
 use std::io::{self, Write as _};
 use std::iter;
-use std::mem;
 use std::panic::{catch_unwind, panic_any, AssertUnwindSafe};
-use std::pin::Pin;
+use std::pin::{pin, Pin};
 use std::sync::{Arc, Mutex, MutexGuard, Weak};
 use std::task::{Context, Poll, RawWaker, RawWakerVTable, Waker};
 
-use futures::future::Map;
 use futures::pin_mut;
 use futures::task::{FutureObj, Spawn, SpawnError};
 use futures::FutureExt as _;
 
 use assert_matches::assert_matches;
 use educe::Educe;
-use itertools::Either;
+use itertools::Either::{self, *};
 use itertools::{chain, izip};
 use slotmap_careful::DenseSlotMap;
 use std::backtrace::Backtrace;
@@ -34,7 +32,7 @@ use strum::EnumIter;
 //    tracing-test = { version = "0.2.4", features = ["no-env-filter"] }
 use tracing::{error, trace};
 
-use oneshot_fused_workaround::{self as oneshot, Canceled, Receiver};
+use oneshot_fused_workaround::{self as oneshot, Canceled};
 use tor_error::error_report;
 use tor_rtcompat::{Blocking, ToplevelBlockOn};
 
@@ -206,26 +204,31 @@ struct Task {
     state: TaskState,
     /// The actual future (or a placeholder for it)
     ///
-    /// May be `None` because we've temporarily moved it out so we can poll it,
+    /// May be `None` briefly in the executor main loop, because we've
+    /// temporarily moved it out so we can poll it,
     /// or if this is a Subthread task which is currently running sync code
     /// (in which case we're blocked in the executor waiting to be
     /// woken up by [`thread_context_switch`](Shared::thread_context_switch).
-    fut: Option<TaskFutureInfo>,
-    /// Is this task actually a [`Subthread`](MockExecutor::subthread_spawn)?
     ///
-    /// Subthread tasks do not end when `fut` is `Ready` -
-    /// instead, `fut` is `Some` when the thread is within `subthread_block_on_future`.
-    /// The rest of the time this is `None`, but we don't run the executor,
-    /// because `Data.thread_to_run` is `ThreadDescriptor::Task(this_task)`.
-    is_subthread: Option<IsSubthread>,
+    /// Note that the `None` can be observed outside the main loop, because
+    /// the main loop unlocks while it polls, so other (non-main-loop) code
+    /// might see it.
+    fut: Option<TaskFutureInfo>,
 }
 
 /// A future as stored in our record of a [`Task`]
+#[derive(Educe)]
+#[educe(Debug)]
 enum TaskFutureInfo {
     /// The [`Future`].  All is normal.
-    Normal(TaskFuture),
+    Normal(#[educe(Debug(ignore))] TaskFuture),
     /// The future isn't here because this task is the main future for `block_on`
     Main,
+    /// This task is actually a [`Subthread`](MockExecutor::subthread_spawn)
+    ///
+    /// Instead of polling it, we'll switch to it with
+    /// [`thread_context_switch`](Shared::thread_context_switch).
+    Subthread,
 }
 
 /// State of a task - do we think it needs to be polled?
@@ -334,6 +337,10 @@ enum ThreadDescriptor {
 #[derive(Copy, Clone, Debug, Eq, PartialEq, Ord, PartialOrd)]
 struct IsSubthread;
 
+/// [`Shared::subthread_yield`] should set our task awake before switching to the executor
+#[derive(Copy, Clone, Debug, Eq, PartialEq, Ord, PartialOrd)]
+struct SetAwake;
+
 thread_local! {
     /// Identifies this thread.
     pub static THREAD_DESCRIPTOR: Cell<ThreadDescriptor> = const {
@@ -417,24 +424,18 @@ impl MockExecutor {
     /// The future passed to `block_on` is not handled here.
     fn spawn_internal(&self, desc: String, fut: TaskFuture) -> TaskId {
         let mut data = self.shared.lock();
-        data.insert_task(desc, TaskFutureInfo::Normal(fut), None)
+        data.insert_task(desc, TaskFutureInfo::Normal(fut))
     }
 }
 
 impl Data {
     /// Insert a task given its `TaskFutureInfo` and return its `TaskId`.
-    fn insert_task(
-        &mut self,
-        desc: String,
-        fut: TaskFutureInfo,
-        is_subthread: Option<IsSubthread>,
-    ) -> TaskId {
+    fn insert_task(&mut self, desc: String, fut: TaskFutureInfo) -> TaskId {
         let state = Awake;
         let id = self.tasks.insert(Task {
             state,
             desc,
             fut: Some(fut),
-            is_subthread,
         });
         self.awake.push_back(id);
         trace!("MockExecutor spawned {:?}={:?}", id, self.tasks[id]);
@@ -449,29 +450,8 @@ impl Spawn for MockExecutor {
     }
 }
 
-impl MockExecutor {
-    /// Implementation of `spawn_blocking` and `blocking_io`
-    fn spawn_thread_inner<F, T>(&self, f: F) -> <Self as Blocking>::ThreadHandle<T>
-    where
-        F: FnOnce() -> T + Send + 'static,
-        T: Send + 'static,
-    {
-        // For the mock executor, everything runs on the same thread.
-        // If we need something more complex in the future, we can change this.
-        let (tx, rx) = oneshot::channel();
-        self.spawn_identified("Blocking".to_string(), async move {
-            match tx.send(f()) {
-                Ok(()) => (),
-                Err(_) => panic!("Failed to send future's output, did future panic?"),
-            }
-        });
-        rx.map(Box::new(|m| m.expect("Failed to receive future's output")))
-    }
-}
-
 impl Blocking for MockExecutor {
-    type ThreadHandle<T: Send + 'static> =
-        Map<Receiver<T>, Box<dyn FnOnce(Result<T, Canceled>) -> T>>;
+    type ThreadHandle<T: Send + 'static> = Pin<Box<dyn Future<Output = T>>>;
 
     fn spawn_blocking<F, T>(&self, f: F) -> Self::ThreadHandle<T>
     where
@@ -483,29 +463,19 @@ impl Blocking for MockExecutor {
             ThreadDescriptor::Executor | ThreadDescriptor::Subthread(_),
  "MockExecutor::spawn_blocking_io only allowed from future or subthread, being run by this executor"
         );
-        self.spawn_thread_inner(f)
+        Box::pin(
+            self.subthread_spawn("spawn_blocking", f)
+                .map(|x| x.expect("Error in spawn_blocking subthread.")),
+        )
     }
 
     fn reenter_block_on<F>(&self, future: F) -> F::Output
     where
-        F: Future + Send + 'static,
+        F: Future,
         F::Output: Send + 'static,
     {
         self.subthread_block_on_future(future)
     }
-
-    fn blocking_io<F, T>(&self, f: F) -> impl Future<Output = T>
-    where
-        F: FnOnce() -> T + Send + 'static,
-        T: Send + 'static,
-    {
-        assert_eq!(
-            THREAD_DESCRIPTOR.get(),
-            ThreadDescriptor::Executor,
-            "MockExecutor::blocking_io only allowed from future being polled by this executor"
-        );
-        self.spawn_thread_inner(f)
-    }
 }
 
 //---------- block_on ----------
@@ -539,7 +509,7 @@ impl ToplevelBlockOn for MockExecutor {
             let main_id = self
                 .shared
                 .lock()
-                .insert_task("main".into(), TaskFutureInfo::Main, None);
+                .insert_task("main".into(), TaskFutureInfo::Main);
             trace!("MockExecutor {main_id:?} is task for block_on");
             self.execute_to_completion(run_store_fut);
         }
@@ -619,13 +589,12 @@ impl MockExecutor {
                 let mut data = self.shared.lock();
 
                 let pus = &mut data.progressing_until_stalled;
-                if let Some(double) = mem::replace(
-                    &mut pus
-                        .as_mut()
-                        .expect("progressing_until_stalled updated under our feet!")
-                        .waker,
-                    Some(waker),
-                ) {
+                if let Some(double) = pus
+                    .as_mut()
+                    .expect("progressing_until_stalled updated under our feet!")
+                    .waker
+                    .replace(waker)
+                {
                     panic!("double progressing_until_stalled.waker! {double:?}");
                 }
 
@@ -690,16 +659,13 @@ impl MockExecutor {
             };
 
             // Poll the selected task
-            let waker = ActualWaker {
-                data: Arc::downgrade(&self.shared),
-                id,
-            }
-            .new_waker();
             trace!("MockExecutor {id:?} polling...");
+            let waker = ActualWaker::make_waker(&self.shared, id);
             let mut cx = Context::from_waker(&waker);
-            let r = match &mut fut {
-                TaskFutureInfo::Normal(fut) => fut.poll_unpin(&mut cx),
-                TaskFutureInfo::Main => main_fut.as_mut().poll(&mut cx),
+            let r: Either<Poll<()>, IsSubthread> = match &mut fut {
+                TaskFutureInfo::Normal(fut) => Left(fut.poll_unpin(&mut cx)),
+                TaskFutureInfo::Main => Left(main_fut.as_mut().poll(&mut cx)),
+                TaskFutureInfo::Subthread => Right(IsSubthread),
             };
 
             // Deal with the returned `Poll`
@@ -711,8 +677,8 @@ impl MockExecutor {
                     .get_mut(id)
                     .expect("task vanished while we were polling it");
 
-                match (r, task.is_subthread) {
-                    (Pending, _) => {
+                match r {
+                    Left(Pending) => {
                         trace!("MockExecutor {id:?} -> Pending");
                         if task.fut.is_some() {
                             panic!("task reinserted while we polled it?!");
@@ -722,7 +688,7 @@ impl MockExecutor {
                         // All we need to do is put the future back.
                         task.fut = Some(fut);
                     }
-                    (Ready(()), None) => {
+                    Left(Ready(())) => {
                         trace!("MockExecutor {id:?} -> Ready");
                         // Oh, it finished!
                         // It might be in `awake`, but that's allowed to contain stale tasks,
@@ -734,15 +700,13 @@ impl MockExecutor {
                         // So, move `fut` to a variable with scope outside the block with `data`.
                         _fut_drop_late = fut;
                     }
-                    (Ready(()), Some(IsSubthread)) => {
+                    Right(IsSubthread) => {
                         trace!("MockExecutor {id:?} -> Ready, waking Subthread");
-                        // Task was blocking on the future given to .subthread_block_on_future().
-                        // That future has completed and stored its result where the Subthread
-                        // can see it.  Now we need to wake up that thread, and let it run
-                        // until it blocks again.
-                        //
-                        // We leave `.fut` as `None`.
-                        // subthread_block_on_future is responsible for filling it in again.
+                        // Task is a subthread, which has called thread_context_switch
+                        // to switch to us.  We "poll" it by switching back.
+
+                        // Put back `TFI::Subthread`, which was moved out temporarily, above.
+                        task.fut = Some(fut);
 
                         self.shared.thread_context_switch(
                             data,
@@ -750,14 +714,10 @@ impl MockExecutor {
                             ThreadDescriptor::Subthread(id),
                         );
 
-                        // Now, if the Subthread still exists, that's because it's waiting
-                        // in subthread_block_on_future again, in which case `fut` is `Some`.
+                        // Now, if the Subthread still exists, that's because it's switched
+                        // back to us, and is waiting in subthread_block_on_future again.
                         // Or it might have ended, in which case it's not in `tasks` any more.
-                        // We can go back to scheduling futures.
-
-                        // `fut` contains the future passed to `subthread_block_on_future`,
-                        // ie it owns an external type.  See above.
-                        _fut_drop_late = fut;
+                        // In any case we can go back to scheduling futures.
                     }
                 }
             }
@@ -801,6 +761,15 @@ impl ActualWaker {
         };
         task.set_awake(self.id, &mut data.awake);
     }
+
+    /// Create and return a `Waker` for task `id`
+    fn make_waker(shared: &Arc<Shared>, id: TaskId) -> Waker {
+        ActualWaker {
+            data: Arc::downgrade(shared),
+            id,
+        }
+        .new_waker()
+    }
 }
 
 //---------- "progress until stalled" functionality ----------
@@ -934,16 +903,7 @@ impl MockExecutor {
 
         {
             let mut data = self.shared.lock();
-            let fut = TaskFutureInfo::Normal(
-                Box::new(
-                    // When the executor decides that this new task is to be polled,
-                    // its future (this future) returns Ready immediately,
-                    // and the executor mainloop will context switch to the new thread.
-                    futures::future::ready(()),
-                )
-                .into(),
-            );
-            let id = data.insert_task(desc.clone(), fut, Some(IsSubthread));
+            let id = data.insert_task(desc.clone(), TaskFutureInfo::Subthread);
 
             let _: std::thread::JoinHandle<()> = std::thread::Builder::new()
                 .name(desc)
@@ -977,20 +937,8 @@ impl MockExecutor {
     ///
     /// If the executor isn't running, `subthread_block_on_future` will hang indefinitely.
     /// See `spawn_subthread`.
-    pub fn subthread_block_on_future<T: Send + 'static>(
-        &self,
-        fut: impl Future<Output = T> + Send + 'static,
-    ) -> T {
-        let ret = Arc::new(Mutex::new(None));
-        let fut = {
-            let ret = ret.clone();
-            async move {
-                let t = fut.await;
-                *ret.lock().expect("poison") = Some(t);
-            }
-        };
-        let fut = TaskFutureInfo::Normal(Box::new(fut).into());
-
+    #[allow(clippy::cognitive_complexity)] // Splitting this up would be worse
+    pub fn subthread_block_on_future<T: Send + 'static>(&self, fut: impl Future<Output = T>) -> T {
         let id = match THREAD_DESCRIPTOR.get() {
             ThreadDescriptor::Subthread(id) => id,
             ThreadDescriptor::Executor => {
@@ -1001,23 +949,38 @@ impl MockExecutor {
             ),
         };
         trace!("MockExecutor thread {id:?}, subthread_block_on_future...");
+        let mut fut = pin!(fut);
 
-        {
-            let mut data = self.shared.lock();
-            let data_ = &mut *data;
-            let task = data_.tasks.get_mut(id).expect("Subthread task vanished!");
-            task.fut = Some(fut);
-            task.set_awake(id, &mut data_.awake);
+        // We yield once before the first poll, and once after Ready, to shake up the
+        // execution order a bit, depending on the scheduling policy.
+        let yield_ = |set_awake| self.shared.subthread_yield(id, set_awake);
+        yield_(Some(SetAwake));
 
-            self.shared.thread_context_switch(
-                data,
-                ThreadDescriptor::Subthread(id),
-                ThreadDescriptor::Executor,
-            );
-        }
+        let ret = loop {
+            // Poll the provided future
+            trace!("MockExecutor thread {id:?}, s.t._block_on_future polling...");
+            let waker = ActualWaker::make_waker(&self.shared, id);
+            let mut cx = Context::from_waker(&waker);
+            let r: Poll<T> = fut.as_mut().poll(&mut cx);
 
-        let ret = ret.lock().expect("poison").take();
-        ret.expect("fut completed but didn't store")
+            if let Ready(r) = r {
+                trace!("MockExecutor thread {id:?}, s.t._block_on_future poll -> Ready");
+                break r;
+            }
+
+            // Pending.  Switch back to the exeuctor thread.
+            // When the future becomes ready, the Waker will be woken, waking the task,
+            // so that the executor will "poll" us again.
+            trace!("MockExecutor thread {id:?}, s.t._block_on_future poll -> Pending");
+
+            yield_(None);
+        };
+
+        yield_(Some(SetAwake));
+
+        trace!("MockExecutor thread {id:?}, subthread_block_on_future complete.");
+
+        ret
     }
 }
 
@@ -1037,8 +1000,8 @@ impl Shared {
         THREAD_DESCRIPTOR.set(ThreadDescriptor::Subthread(id));
         trace!("MockExecutor thread {id:?}, entrypoint");
 
-        // Wait for the executor to tell us to run.
-        // This will be done the first time the task is polled.
+        // We start out Awake, but we wait for the executor to tell us to run.
+        // This will be done the first time the task is "polled".
         {
             let data = self.lock();
             self.thread_context_switch_waitfor_instruction_to_run(
@@ -1055,7 +1018,7 @@ impl Shared {
 
         trace!("MockExecutor thread {id:?}, completed user code");
 
-        // This makes SubthreadFuture ready.
+        // This makes the return value from subthread_spawn ready.
         // It will be polled by the executor in due course, presumably.
 
         output_tx.send(ret).unwrap_or_else(
@@ -1079,6 +1042,37 @@ impl Shared {
         }
     }
 
+    /// Yield back to the executor from a subthread
+    ///
+    /// Checks that things are in order
+    /// (in particular, that this task is in the data structure as a subhtread)
+    /// and switches to the executor thread.
+    ///
+    /// The caller must arrange that the task gets woken.
+    ///
+    /// With [`SetAwake`], sets our task awake, so that we'll be polled
+    /// again as soon as we get to the top of the executor's queue.
+    /// Otherwise, we'll be reentered after someone wakes a [`Waker`] for the task.
+    fn subthread_yield(&self, us: TaskId, set_awake: Option<SetAwake>) {
+        let mut data = self.lock();
+        {
+            let data = &mut *data;
+            let task = data.tasks.get_mut(us).expect("Subthread task vanished!");
+            match &task.fut {
+                Some(TaskFutureInfo::Subthread) => {}
+                other => panic!("subthread_block_on_future but TFI {other:?}"),
+            };
+            if let Some(SetAwake) = set_awake {
+                task.set_awake(us, &mut data.awake);
+            }
+        }
+        self.thread_context_switch(
+            data,
+            ThreadDescriptor::Subthread(us),
+            ThreadDescriptor::Executor,
+        );
+    }
+
     /// Switch from (sub)thread `us` to (sub)thread `them`
     ///
     /// Returns when someone calls `thread_context_switch(.., us)`.
@@ -1397,14 +1391,14 @@ impl MockExecutor {
     /// see [`.debug_dump()`](MockExecutor::debug_dump).
     pub fn as_debug_dump(&self) -> DebugDump {
         let data = self.shared.lock();
-        DebugDump(Either::Right(data))
+        DebugDump(Right(data))
     }
 }
 
 impl Data {
     /// Convenience function: dump including backtraces, to stderr
     fn debug_dump(&mut self) {
-        DebugDump(Either::Left(self)).to_stderr();
+        DebugDump(Left(self)).to_stderr();
     }
 }
 
@@ -1434,22 +1428,14 @@ impl Debug for DebugDump<'_> {
 // See `impl Debug for Data` for notes on the output
 impl Debug for Task {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let Task {
-            desc,
-            state,
-            fut,
-            is_subthread,
-        } = self;
+        let Task { desc, state, fut } = self;
         write!(f, "{:?}", desc)?;
         write!(f, "=")?;
-        match is_subthread {
-            None => {}
-            Some(IsSubthread) => write!(f, "T")?,
-        }
         match fut {
             None => write!(f, "P")?,
             Some(TaskFutureInfo::Normal(_)) => write!(f, "f")?,
             Some(TaskFutureInfo::Main) => write!(f, "m")?,
+            Some(TaskFutureInfo::Subthread) => write!(f, "T")?,
         }
         match state {
             Awake => write!(f, "W")?,
@@ -1703,7 +1689,7 @@ mod test {
 
     #[cfg_attr(not(miri), traced_test)]
     #[test]
-    fn subthread() {
+    fn subthread_oneshot() {
         for runtime in various_mock_executors() {
             runtime.block_on(async {
                 let (tx, rx) = oneshot::channel();
@@ -1726,4 +1712,47 @@ mod test {
             });
         }
     }
+
+    #[cfg_attr(not(miri), traced_test)]
+    #[test]
+    #[allow(clippy::cognitive_complexity)] // It's is not that complicated, really.
+    fn subthread_pingpong() {
+        for runtime in various_mock_executors() {
+            runtime.block_on(async {
+                let (mut i_tx, mut i_rx) = mpsc::channel(1);
+                let (mut o_tx, mut o_rx) = mpsc::channel(1);
+                info!("spawning subthread");
+                let thr = runtime.subthread_spawn("thr", {
+                    let runtime = runtime.clone();
+                    move || {
+                        while let Some(i) = {
+                            info!("thread receiving ...");
+                            runtime.subthread_block_on_future(i_rx.next())
+                        } {
+                            let o = i + 12;
+                            info!("thread received {i}, sending {o}");
+                            runtime.subthread_block_on_future(o_tx.send(o)).unwrap();
+                            info!("thread sent {o}");
+                        }
+                        info!("thread exiting");
+                        42
+                    }
+                });
+                for i in 0..2 {
+                    info!("main task sending {i}");
+                    i_tx.send(i).await.unwrap();
+                    info!("main task sent {i}");
+                    let o = o_rx.next().await.unwrap();
+                    info!("main task recv => {o}");
+                    assert_eq!(o, i + 12);
+                }
+                info!("main task dropping sender");
+                drop(i_tx);
+                info!("main task awaiting thread");
+                let r = thr.await.unwrap();
+                info!("main task complete");
+                assert_eq!(r, 42);
+            });
+        }
+    }
 }
